from datetime import datetime, timezone
from unittest import TestCase

from coordinates_label_photos.coordinates import Coordinates
from coordinates_label_photos.parsers import coords_parser


class TestCoordinatesCollection(TestCase):
    filename = 'resources/photo-gps/track.gpx'
    coords_collection = coords_parser(filename)

    def test_start_time(self):
        got = self.coords_collection.start_time()
        self.assertEqual(datetime(2022, 3, 6, 10, 28, 49, tzinfo=timezone.utc), got)

    def test_end_time(self):
        got = self.coords_collection.end_time()
        self.assertEqual(datetime(2022, 3, 6, 10, 30, 13, tzinfo=timezone.utc), got)

    def test_interpolate_position_before(self):
        timestamp = datetime(2022, 3, 6, 10, 28, 0, tzinfo=timezone.utc)

        got = self.coords_collection.interpolate_position(timestamp)

        self.assertEqual(None, got)

    def test_interpolate_position_after(self):
        timestamp = datetime(2022, 3, 6, 10, 31, 0, tzinfo=timezone.utc)

        got = self.coords_collection.interpolate_position(timestamp)

        self.assertEqual(None, got)

    def test_interpolate_position_exact(self):
        timestamp = datetime(2022, 3, 6, 10, 30, 9, tzinfo=timezone.utc)

        got = self.coords_collection.interpolate_position(timestamp)

        expected = Coordinates(
            lat=46.119549,
            lon=7.021491,
            elevation=938.0,
            timestamp=timestamp
        )

        self.assertAlmostEqual(expected.lat, got.lat, 6)
        self.assertAlmostEqual(expected.lon, got.lon, 6)
        self.assertAlmostEqual(expected.elevation, got.elevation, 6)

        self.assertEqual(timestamp, got.timestamp)

    def test_interpolate_position_middle(self):
        timestamp = datetime(2022, 3, 6, 10, 30, 9, microsecond=250000, tzinfo=timezone.utc)
        got = self.coords_collection.interpolate_position(timestamp)

        expected = Coordinates(
            lat=46.119548,
            lon=7.021491,
            elevation=938.0,
            timestamp=timestamp
        )

        self.assertAlmostEqual(expected.lat, got.lat, 6)
        self.assertAlmostEqual(expected.lon, got.lon, 6)
        self.assertAlmostEqual(expected.elevation, got.elevation, 6)

        self.assertEqual(timestamp, got.timestamp)

    def test_lat_lon_boundaries(self):
        got = self.coords_collection.lat_lon_boundaries()

        self.assertEqual(46.11961992457509, got[0].lat)
        self.assertEqual(7.021400593221188, got[0].lon)
        self.assertEqual(46.11953526735306, got[1].lat)
        self.assertEqual(7.021564207971096, got[1].lon)




from datetime import timezone, datetime
from unittest import TestCase

from coordinates_label_photos.parsers import emlid_fwf_line_parser


class TestParsers(TestCase):
    def test_emlid_llh_line_parser(self):
        given = '2022/08/03 07:35:30.600   37.413804352   23.350134190    44.1240   2  11   0.2400   ' \
                '0.2700   0.8700   0.0000   0.0000   0.0000   4.60    0.0'

        got = emlid_fwf_line_parser(given)

        self.assertEqual(datetime(2022, 8, 3, 7, 35, 30, 600000, tzinfo=timezone.utc), got.timestamp)
        self.assertEqual(37.413804352, got.lat)
        self.assertEqual(23.350134190, got.lon)
        self.assertEqual(44.1240, got.elevation)

    def test_pos_line_parser(self):
        given = '2022/08/08 05:24:20.200   37.428194740   23.134118977    39.6114   2  16   0.8445   0.6919   ' \
                '2.4424  -0.1155   0.5825   0.8769   0.00    0.0'

        got = emlid_fwf_line_parser(given)

        self.assertEqual(datetime(2022, 8, 8, 5, 24, 20, 200000, tzinfo=timezone.utc), got.timestamp)
        self.assertEqual(37.428194740, got.lat)
        self.assertEqual(23.134118977, got.lon)
        self.assertEqual(39.6114, got.elevation)
        self.assertAlmostEqual(1.65522, got.horiz_accuracy, 8)
        self.assertAlmostEqual(4.787104, got.vert_accuracy, 8)



from pyproj import CRS, Transformer, crs, Geod, Proj
from pyproj.aoi import AreaOfInterest
from pyproj.database import query_utm_crs_info

from coordinates_label_photos.coordinates.coordinates_collection import CoordinatesCollection


def crs_epsg_code(coords: CoordinatesCollection) -> str:
    (nw, se) = coords.lat_lon_boundaries()
    utm_crs_list = query_utm_crs_info(
        datum_name="WGS 84",
        area_of_interest=AreaOfInterest(
            west_lon_degree=nw.lon,
            south_lat_degree=se.lat,
            east_lon_degree=se.lon,
            north_lat_degree=nw.lat,
        ),
    )
    return utm_crs_list[0].code


def crs_wgs84_utm_tile(coords: CoordinatesCollection):
    """

    :param coords:
    :type coords:
    :return:
    :rtype:
    """
    name = CRS.from_epsg(crs_epsg_code(coords)).name
    return name.replace('WGS 84 / UTM zone ', 'WGS84 UTM ')


def crs_transformer_deg_utm(epsg_code: str) -> Transformer:
    """
    From a given coords collection, gives the transformation function long/lat -> x/y
    :param epsg_code: the area EPSG code
    :type epsg_code: str
    :return: a pyproj transformer object
    :rtype: pyproj.Transformer
    """
    crs_latlon = CRS.from_epsg('4326') # WGS84
    crs = CRS.from_epsg(epsg_code)
    return Transformer.from_crs(crs_latlon.geodetic_crs, crs, always_xy=True)



import argparse
import logging

from coordinates_label_photos.coordinates.crs_transformer import crs_epsg_code, crs_transformer_deg_utm
from coordinates_label_photos.parsers import coords_parser
from coordinates_label_photos.photos import list_photo_filenames, calibrate_photo
from coordinates_label_photos.reports.coordinates_collection import report_image_coordinates_collections


def main():
    logging.basicConfig(level=logging.INFO)
    parser = argparse.ArgumentParser(description='Labels photos with a GPS track points (.gpx)')

    parser.add_argument(
        "--coords",
        help=".gpx|.llh coordinates file",
        type=str,
        required=True,
        metavar='/path/to/coords'
    )
    parser.add_argument(
        "--images",
        help="image directory (edited in place) or glob",
        type=str,
        required=True,
        metavar='/path/to/images/*/*.JPG'
    )
    parser.add_argument(
        "--images-timestamp-offset",
        help="In case it0's not in the photo EXIF (e.g. '+3:00)",
        type=str,
        required=False,
        metavar='+3:00'
    )
    parser.add_argument(
        "--odm-geo",
        help="Image locations in a OpenDroneMap geo location format",
        type=str,
        metavar='/path/to/odm-geo.txt'
    )
    parser.add_argument(
        "--epsg-code",
        help="EPSG digit code. If none provided, it will be guessed from the coordinates",
        type=str,
        metavar='22332'
    )
    parser.add_argument(
        "--horizontal-accuracy",
        help="horizontal accuracy (GPS + device) in meters",
        type=float,
        metavar='0.4'
    )
    parser.add_argument(
        "--vertical-accuracy",
        help="vertical accuracy (GPS + device) in meters",
        type=float,
        metavar='0.3'
    )
    parser.add_argument(
        "--camera-vertical-offset",
        help="the difference, in altitude, between the camera and the GPS," +
             " in meters. Number is negative if the camera is below the GPS",
        type=float,
        metavar='0.52'
    )
    parser.add_argument(
        "--skip-photo-edit",
        help="When not set, the photos are edited in place",
        action='store_true'
    )
    parser.add_argument(
        "--report-photo-locations",
        help="an image filename where to report the photo locations",
        type=str,
        metavar='/path/to/report-photo-locations.jpeg'
    )
    parser.add_argument(
        "--report-track",
        help="an image filename where to report the original GPS coordinates",
        type=str,
        metavar='/path/to/report-track.jpeg'
    )

    args = parser.parse_args()

    photos = list_photo_filenames(getattr(args, 'images'))
    track_coords = coords_parser(args.coords)
    photo_coords = calibrate_photo(
        photos=photos,
        track_coords=track_coords,
        default_time_offset=args.images_timestamp_offset,
        camera_vertical_offset=args.camera_vertical_offset,
        horizontal_accuracy=args.horizontal_accuracy,
        vertical_accuracy=args.vertical_accuracy,
        edit_photo=not args.skip_photo_edit
    )

    if args.odm_geo:
        logging.info('Saving image locations in %s' % args.odm_geo)
        if args.epsg_code is not None:
            epsg_code = args.epsg_code
        else:
            epsg_code = crs_epsg_code(track_coords)
        csr_transfo = crs_transformer_deg_utm(epsg_code)
        with open(args.odm_geo, 'w') as fd_geo:
            fd_geo.write('EPSG:%s\n' % epsg_code)
            for c in photo_coords.points:
                (x, y) = csr_transfo.transform(c.lon, c.lat)
                horiz_accuracy = c.horiz_accuracy
                if horiz_accuracy is None:
                    horiz_accuracy = 0
                vert_accuracy = c.vert_accuracy
                if vert_accuracy is None:
                    vert_accuracy = 0
                line = '%s\t%.10f\t%.10f\t%.3f\t0\t0\t0\t%.3f\t%.3f\n' % (
                    c.label,
                    x,
                    y,
                    c.elevation,
                    horiz_accuracy,
                    vert_accuracy
                )
                fd_geo.write(line)

    if getattr(args, 'report_track') is not None:
        logging.info('Saving track %s' % getattr(args, 'report_track'))
        report_image_coordinates_collections(track_coords, getattr(args, 'report_track'), 1000)
    if getattr(args, 'report_photo_locations') is not None:
        report_image_coordinates_collections(photo_coords, getattr(args, 'report_photo_locations'), 1000)


if __name__ == '__main__':
    main()



from unittest import TestCase

from coordinates_label_photos.coordinates.crs_transformer import crs_wgs84_utm_tile, crs_epsg_code, \
    crs_transformer_deg_utm
from coordinates_label_photos.parsers import coords_parser


class TestCrsTransformation(TestCase):
    filename = 'resources/photo-gps/track.gpx'
    coords_collection = coords_parser(filename)

    def test_epsg(self):
        got = crs_epsg_code(self.coords_collection)

        self.assertEqual('32632', got)

    def test_wgs84_utm_zone(self):
        got = crs_wgs84_utm_tile(self.coords_collection)

        self.assertEqual('WGS84 UTM 32N', got)

    def test_pyproj_transformer_deg_utm_1(self):
        transfo = crs_transformer_deg_utm('22332')
        (got_x, got_y) = transfo.transform(23.13361000, 37.42807200)
        self.assertAlmostEqual(1753779.712, got_x, 2)
        self.assertAlmostEqual(4237153.604, got_y, 2)

    def test_pyproj_transformer_deg_utm_22332(self):
        # ref https://epsg.io/transform#s_srs=4326&t_srs=22332&x=23.1336358&y=37.4280617
        transfo = crs_transformer_deg_utm('22332')
        (got_x, got_y) = transfo.transform(23.1336358, 37.4280617)
        self.assertAlmostEqual(1753697.3216273494, got_x, 2)
        self.assertAlmostEqual(4236984.973854066, got_y, 2)

    def test_pyproj_transformer_deg_utm_32632(self):
        # checked with
        # * https://epsg.io/transform#s_srs=4326&t_srs=22332&x=23.1336358&y=37.4280617
        # * https://mygeodata.cloud/cs2cs/
        transfo = crs_transformer_deg_utm('32632')
        (got_x, got_y) = transfo.transform(23.1336358, 37.4280617)
        self.assertAlmostEqual(1753734.2744281862, got_x, 2)
        self.assertAlmostEqual(4237442.419149317, got_y, 2)



import argparse
import logging
import os
import csv
import shutil
from pathlib import Path

from coordinates_label_photos.coordinates import exif_to_coordinates, Coordinates
from coordinates_label_photos.coordinates.coordinates_collection import CoordinatesCollection
from coordinates_label_photos.photos import list_photo_filenames, get_photo_exif
from PIL import Image, ImageDraw, ImageFont


def load_coords_csv(filename) -> CoordinatesCollection:
    logging.info('Loading CSV coordinates from %s' % filename)
    with open(filename, newline='') as csvfile:
        reader = csv.DictReader(csvfile)
        coords = []

        for row in reader:
            coords.append(
                Coordinates(
                    lat=float(row['Latitude']),
                    lon=float(row['Longitude']),
                    elevation=float(row['Elevation']),
                    timestamp=None,
                    label=row['Name']
                )
            )
        collect = CoordinatesCollection(coords)
        logging.info('Loaded CSV with %d track points between %s and %s' % (
            len(collect), collect.start_time(), collect.end_time()))
        return collect


def report(
        all_photos,
        selected_photos,
        coords,
        filename,
        max_dim
):
    logging.info('Saving report in %s' % filename)
    all_coords = [c for c in coords.points]
    all_coords.extend(all_photos.values())
    all_coords = CoordinatesCollection(all_coords)

    boundaries = all_coords.lat_lon_boundaries()
    dim_lat = boundaries[1].lat - boundaries[0].lat
    dim_lon = boundaries[1].lon - boundaries[0].lon
    dimensions = all_coords.dimensions()
    if dimensions[0] > dimensions[1]:
        width = max_dim * 0.8
        height = max_dim * dimensions[1] / dimensions[0]
    else:
        height = max_dim * 0.8
        width = max_dim * dimensions[1] / dimensions[0]

    margin = max_dim * 0.1

    img = Image.new('RGB', (int(width + 2 * margin), int(height + 2 * margin)), 'white')
    draw = ImageDraw.Draw(img)
    font = ImageFont.truetype('/System/Library/Fonts/Geneva.ttf', size=10)

    def coord_xy(c: Coordinates):
        (x, y) = (margin + int((c.lon - boundaries[0].lon) / dim_lon * width),
                  margin + int((c.lat - boundaries[0].lat) / dim_lat * height))
        # print('%s\t%d\t%d' % (c, x, y))

        return x, y

    prev = None
    for c in coords.points:
        (x, y) = coord_xy(c)
        draw.text((x + 4, y + 4), c.label, fill='black', font=font)
        draw.ellipse((x - 2, y - 2, x + 2, y + 2), fill='black')

        if prev is not None:
            draw.line([(x, y), (prev['x'], prev['y'])], fill='grey')
        prev = {
            'x': x,
            'y': y,
        }
    for (p, c) in all_photos.items():
        (x, y) = coord_xy(c)
        name = Path(p).stem
        draw.text((x + 4, y + 4), name, fill='black', font=font)
        draw.ellipse((x - 2, y - 2, x + 2, y + 2), fill='red')

    for (p, c) in selected_photos.items():
        (x, y) = coord_xy(all_photos[p])
        draw.ellipse((x - 2, y - 2, x + 2, y + 2), fill='green')
        (xc, yc) = coord_xy(c)
        draw.line([(x, y), (xc, yc)], fill='green')

    img.save(filename)


def main():
    logging.basicConfig(level=logging.DEBUG)
    parser = argparse.ArgumentParser(description='Select the closest pictures to a given list of positions')

    parser.add_argument(
        "--coords",
        help='emlid csv coordinate files',
        type=str,
        metavar='path/to/coords.csv'

    )
    parser.add_argument(
        "--images",
        help="image directory",
        type=str,
        required=True,
        metavar='/path/to/images'
    )
    parser.add_argument(
        "--dest",
        help="the destination directory for the selected images",
        type=str,
        required=True,
        metavar='/path/to/dir'
    )
    parser.add_argument(
        "--report",
        help="an image filename where to report matching photos",
        type=str,
        metavar='/path/to/report-photo-closest.jpeg'
    )
    args = parser.parse_args()

    photo_coords = {}
    for photo in list_photo_filenames(args.images):
        exif = get_photo_exif(photo)
        c = exif_to_coordinates(exif)
        c.elevation = 0
        photo_coords[photo] = c

    selected_photos = {}
    coords = load_coords_csv(args.coords)
    for coord in coords.points:
        # find closest
        closest_photo = min(photo_coords.keys(), key=lambda p: coord.distance(photo_coords[p]))

        if closest_photo not in selected_photos:
            # it must be the closest photo to the point, but the reciprocal must also be true
            closest_coords_to_photo = min(coords.points, key=lambda c1: c1.distance(photo_coords[closest_photo]))
            if closest_coords_to_photo == coord:
                selected_photos[closest_photo] = coord

    dest_dir = args.dest
    os.makedirs(dest_dir, exist_ok=True)
    for photo in selected_photos.keys():
        shutil.copy(photo, dest_dir)

    if args.report is not None:
        report(
            all_photos=photo_coords,
            selected_photos=selected_photos,
            coords=coords,
            max_dim=2500,
            filename=args.report
        )


if __name__ == '__main__':
    main()



from datetime import datetime, timezone
from unittest import TestCase

from coordinates_label_photos.photos import get_photo_timestamp


class TestPhoto(TestCase):
    def test_get_photo_timestamp(self):
        # BTW, the picture shows a watch displaying 11:29:10, but the EXIF tag is 11:29:11
        got = get_photo_timestamp('resources/photo-gps/IMG_4967.jpeg')

        self.assertEqual(datetime(2022, 3, 6, 10, 29, 11, tzinfo=timezone.utc), got)



from . import to_abs_deg_min_sec, Coordinates
from unittest import TestCase


class TestCoordinates(TestCase):
    coords = Coordinates(-12.34567, 98.7654)

    def test_to_abs_deg_min_sec(self):
        got = to_abs_deg_min_sec(self.coords.lat)

        self.assertEqual(got, (12, 20, 44.412))

    def test_exiv_lat(self):
        self.assertEqual('S', self.coords.lat_dir())
        self.assertEqual([(12, 1), (20, 1), (11103, 250)], self.coords.exiv_lat())

    def test_exiv_lon(self):
        self.assertEqual('W', self.coords.lon_dir())
        self.assertEqual([(98, 1), (45, 1), (1386, 25)], self.coords.exiv_lon())

    def test_distance(self):
        c1 = Coordinates(52.2296756, 21.0122287)
        c2 = Coordinates(52.406374, 16.9251681)

        got = c1.distance(c2)

        self.assertAlmostEqual(279352.9, got, 1)



from setuptools import setup

setup(
    entry_points={
        'console_scripts': [
            'coordinates-label-photos=coordinates_label_photos.scripts.label_photos:main',
            'coordinates-clear-photos=coordinates_label_photos.scripts.clear_photo_coords:main',
            'plot-coords-tracks=coordinates_label_photos.scripts.plot_coords_tracks:main',
            'images-to-gpx=coordinates_label_photos.scripts.images_to_gpx:main',
            'select-closest-photos=coordinates_label_photos.scripts.select_closest_photos:main',
        ],
    }
)
